package tudelft.mocktutorial;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.InOrder;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.LinkedList;
import java.util.List;

import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.atLeastOnce;
import static org.mockito.Mockito.atMost;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.inOrder;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.verifyZeroInteractions;
import static org.mockito.Mockito.when;

public class MockingTutorialTest {
    @Mock(name = "myList")
    LinkedList mockedList;


    @BeforeEach
    public void setUp() {
        //You can mock concrete classes, not just interfaces
        //mockedList = mock(LinkedList.class, "mockedList");
        MockitoAnnotations.initMocks(this);
    }

    private void initializeListForVerification() {
        mockedList.add("once");

        mockedList.add("twice");
        mockedList.add("twice");

        mockedList.add("three times");
        mockedList.add("three times");
        mockedList.add("three times");
    }

    /**
     * Test method to show simple mocking
     */
    @Test
    void mockedList() {
        //using mock object
        mockedList.add("one");
        mockedList.clear();

        //verification
        verify(mockedList).add("one");
        verify(mockedList).clear();
    }

    /**
     * Test method to show stubbing with Mockito
     */
    @Test
    public void stubbedList() {

        //stubbing
        //when(mockedList.get(anyInt())).thenReturn("first");
        when(mockedList.get(anyInt())).thenReturn("FIRST");
        //when(mockedList.get(1)).thenThrow(new RuntimeException()).thenReturn("Ouch!");

        //following prints "first"
        System.out.println(mockedList.get(0));

        //following throws runtime exception
        System.out.println(mockedList.get(1));

        //following prints "null" because get(999) was not stubbed
        System.out.println(mockedList.get(999));

        //Although it is possible to verify a stubbed invocation, usually it's just redundant
        //If your code cares what get(0) returns, then something else breaks (often even before verify() gets executed).
        //If your code doesn't care what get(0) returns, then it should not be stubbed.
        //verify(mockedList).get(1);

    }

    @Test
    void argumentMatcherExample() {
        //stubbing using built-in anyInt() argument matcher
        when(mockedList.get(anyInt())).thenReturn("element");
        when(mockedList.get(0)).thenReturn("first");

        //stubbing using custom matcher (let's say isValid() returns your own matcher implementation):
       // when(mockedList.contains(argThat(isValid()))).thenReturn("element");

        //following prints "element"
        System.out.println(mockedList.get(0));
        System.out.println(mockedList.get(999));

        //you can also verify using an argument matcher
        //verify(mockedList).get(anyInt());
    }

    @Test
    void verifyExample() {
        initializeListForVerification();

        //following two verifications work exactly the same - times(1) is used by default
        verify(mockedList).add("once");
        verify(mockedList, times(2)).add("twice");
    }


    @Test
    void verifyMultipleTimesAndNever() {
        initializeListForVerification();

        //exact number of invocations verification
        verify(mockedList, times(2)).add("twice");
        verify(mockedList, times(3)).add("three times");

        //verification using never(). never() is an alias to times(0)
        verify(mockedList, never()).add("never happened");
    }

    @Test
    void verifyAtMostAtLeast() {
        initializeListForVerification();

        //verification using atLeast()/atMost()
        verify(mockedList, atMost(1)).add("once");
        verify(mockedList, atLeastOnce()).add("three times");
        verify(mockedList, atLeast(2)).add("three times");
        verify(mockedList, atMost(5)).add("three times");
    }

    @Test
    void orderSingleMock() {
        List singleMock = mock(List.class, "orderList");

        //using a single mock
        singleMock.add("was added first");
        singleMock.add("was added second");


        //create an inOrder verifier for a single mock
        InOrder inOrder = inOrder(singleMock);

        //following will make sure that add is first called with "was added first", then with "was added second"
        inOrder.verify(singleMock).add("was added first");
        inOrder.verify(singleMock).add("was added second");
    }

    @Test
    void orderDoubleMock() {
// B. Multiple mocks that must be used in a particular order
        List firstMock = mock(List.class);
        List secondMock = mock(List.class);

        //using mocks
        firstMock.add("was called first");
        secondMock.add("was called second");

        //create inOrder object passing any mocks that need to be verified in order
        InOrder inOrder = inOrder(firstMock, secondMock);

        //following will make sure that firstMock was called before secondMock
        inOrder.verify(firstMock).add("was called first");
        inOrder.verify(secondMock).add("was called second");
    }

    @Test
    void testNoInteractions() {
        List firstMock = mock(List.class);
        List secondMock = mock(List.class);
        //using mocks - only mockOne is interacted
        mockedList.add("one");
        firstMock.add("two");

        //ordinary verification
        verify(mockedList).add("one");

        //verify that method was never called on a mock
       // verify(mockedList, never()).add("two");
        //firstMock.add("Foo");

        //verify that other mocks were not interacted
        verifyZeroInteractions(firstMock, secondMock);
        verifyNoMoreInteractions(firstMock);
    }

    @Test
    void testRedundantInteractions() {
        //using mocks
        mockedList.add("one");
        mockedList.add("two");
        //mockedList.getFirst();

        verify(mockedList).add("one");
        verify(mockedList).add("two");


        //following verification will fail
        verifyNoMoreInteractions(mockedList);
    }

    @Test
    void spyTest() {
        List list = new LinkedList();
        List spy = spy(list);

        //optionally, you can stub out some methods:
        when(spy.size()).thenReturn(100);

        //using the spy calls *real* methods
        spy.add("one");
        spy.add("two");

        //prints "one" - the first element of a list
        System.out.println(spy.get(0));

        //size() method was stubbed - 100 is printed
        System.out.println(spy.size());

        //optionally, you can verify
        verify(spy).add("one");
        verify(spy).add("two");
    }

    @Test
    void alternativeStubTest() {
        List list = new LinkedList();
        List spy = spy(list);

        //Impossible: real method is called so spy.get(0) throws IndexOutOfBoundsException (the list is yet empty)
        when(spy.get(0)).thenReturn("foo");

        //You have to use doReturn() for stubbing
        doReturn("foo").when(spy).get(0);
    }
}
